from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_two_alternatives():
    regexEnumerator = RegexEnumerator(r'a|b')
    possibilities = ['a', 'b']

    f_finite(regexEnumerator, possibilities)


def test_alternatives_with_quantifier_on_second_option():
    regexEnumerator = RegexEnumerator(r'a|b*')
    possibilities = ['a', '', 'b', 'bb', 'bbb', 'bbbb', 'bbbbb']

    f_infinite(regexEnumerator, possibilities)


def test_alternatives_with_quantifier_plus_on_first_option():
    regexEnumerator = RegexEnumerator(r'a+|b')
    possibilities = ['b', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)


def test_multiple_alternatives():
    regexEnumerator = RegexEnumerator(r'a|b|c')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)


def test_alternative_with_literal_and_character_class():
    regexEnumerator = RegexEnumerator(r'a|[b-d]')
    possibilities = ['a', 'b', 'c', 'd']

    f_finite(regexEnumerator, possibilities)


def test_alternative_with_character_class_and_literal():
    regexEnumerator = RegexEnumerator(r'[a-c]{ 0}|d')
    possibilities = ['', 'd']

    f_finite(regexEnumerator, possibilities)


def test_alternation_with_character_classes_and_literals():
    regexEnumerator = RegexEnumerator(r'(a|[0-2])')
    possibilities = ['a', '0', '1', '2']

    f_finite(regexEnumerator, possibilities)


def test_nested_alternation():
    regexEnumerator = RegexEnumerator(r'((a|b)|c)')
    possibilities = ['a', 'b', 'c']
    
    f_finite(regexEnumerator, possibilities)


def test_alternation_with_grouping():
    regexEnumerator = RegexEnumerator(r'(a(b|c)d|x)')
    possibilities = ['abd', 'acd', 'x']
    
    f_finite(regexEnumerator, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_empty_pattern_yields_empty_string():
    regexEnumerator = RegexEnumerator(r'')
    possibilities = ['']
    f_finite(regexEnumerator, possibilities)


def test_single_literal_character():
    regexEnumerator = RegexEnumerator(r'a')
    possibilities = ['a']
    f_finite(regexEnumerator, possibilities)


def test_zero_or_more_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regexEnumerator, possibilities)


def test_one_or_more_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a+')
    possibilities = ['a', 'aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regexEnumerator, possibilities)


def test_zero_or_one_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a?')
    possibilities = ['', 'a']
    f_finite(regexEnumerator, possibilities)


def test_exact_repetition_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a{2}')
    possibilities = ['aa']
    f_finite(regexEnumerator, possibilities)


def test_minimum_repetition_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a{2,}')
    possibilities = ['aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regexEnumerator, possibilities)


def test_min_max_repetition_quantifier_on_single_char():
    # `a{2,4}` yields 'aa', 'aaa', 'aaaa'.
    regexEnumerator = RegexEnumerator(r'a{2,4}')
    possibilities = ['aa', 'aaa', 'aaaa']
    f_finite(regexEnumerator, possibilities)


def test_zero_times_repetition_quantifier_on_single_char():
    regexEnumerator = RegexEnumerator(r'a{0}')
    possibilities = ['']
    f_finite(regexEnumerator, possibilities)


def test_escaped_literal_special_characters():
    regexEnumerator = RegexEnumerator(r'\*\+\?')
    possibilities = ['*+?']
    f_finite(regexEnumerator, possibilities)


def test_single_character_class():
    regexEnumerator = RegexEnumerator(r'[abc]')
    possibilities = ['a', 'b', 'c']
    f_finite(regexEnumerator, possibilities)


def test_single_escaped_character():
    regexEnumerator = RegexEnumerator(r'\n')
    possibilities = ['\n']
    f_finite(regexEnumerator, possibilities)


def test_literal_dot_character():
    regexEnumerator = RegexEnumerator(r'\.')
    possibilities = ['.']
    f_finite(regexEnumerator, possibilities)



from regex_enumerator import RegexEnumerator


def f_finite(regexEnumerator: RegexEnumerator, possibilities: list[str]):
    while len(possibilities) != 0:
        res = regexEnumerator.next()
        assert res in possibilities, f"'{res}' is not in the possibilities."
        possibilities.remove(res)

    assert regexEnumerator.done


def f_infinite(regexEnumerator: RegexEnumerator, possibilities: list[str]):
    while len(possibilities) != 0:
        res = regexEnumerator.next()
        assert res in possibilities, f"'{res}' is not in the possibilities."
        possibilities.remove(res)

    assert not regexEnumerator.done



from .regex_parser import RegexParser
from .regex_tree import RegexTree


class RegexEnumerator:
    def __init__(self, regex: str, additional_charset: str | list[str] = None) -> None:
        default_charset = [chr(c) for c in range(32, 127)]

        if additional_charset is None:
            additional = []
        elif isinstance(additional_charset, list):
            additional = list(''.join(additional_charset))
        else:
            additional = list(additional_charset)

        charset = ''.join(sorted(set(default_charset + additional)))
        parser = RegexParser(regex, charset)
        self.regexTree: RegexTree = parser.parse()
        self.current: list[str] = list(self.regexTree.current)
        self.done: bool = self.regexTree.done and len(self.current) == 0

    def next(self) -> str | None:
        if len(self.current) != 0:
            res = self.current.pop()
            self.done = self.regexTree.done and len(self.current) == 0
            return res

        while True:
            if self.regexTree.done:
                self.done = True
                return None
            self.current = list(self.regexTree.next())
            if len(self.current) != 0:
                break

        res = self.current.pop()
        self.done = self.regexTree.done and len(self.current) == 0
        return res



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_not_capturing_groups():
    regexEnumerator = RegexEnumerator(r'(?:a)(b)\1')
    possibilities = ['abb']

    f_finite(regexEnumerator, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_single_character_class():
    regexEnumerator = RegexEnumerator(r'[a]')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)


def test_character_class_with_two_literals():
    regexEnumerator = RegexEnumerator(r'[ab]')
    possibilities = ['a', 'b']

    f_finite(regexEnumerator, possibilities)


def test_character_class_with_zero_or_more_quantifier():
    regexEnumerator = RegexEnumerator(r'[a]*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)


def test_range_character_class():
    regexEnumerator = RegexEnumerator(r'[a-c]')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)


def test_range_character_class_with_repetition():
    regexEnumerator = RegexEnumerator(r'[a-c]{1,2}')
    possibilities = ['a', 'b', 'c', 'aa', 'ab',
                     'ac', 'ba', 'bb', 'bc', 'ca', 'cb', 'cc']

    f_finite(regexEnumerator, possibilities)


def test_range_character_class_with_zero_repetition():
    regexEnumerator = RegexEnumerator(r'[a-c]{0}')
    possibilities = ['']

    f_finite(regexEnumerator, possibilities)


def test_range_character_class_with_one_or_more_quantifier():
    regexEnumerator = RegexEnumerator(r'[a-b]+')
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb', 'aaa',
                     'aab', 'aba', 'abb', 'baa', 'bab', 'bba', 'bbb']

    f_infinite(regexEnumerator, possibilities)


def test_two_ranges_with_optional_quantifier():
    regexEnumerator = RegexEnumerator(r'[a-cf-g]?')
    possibilities = ['', 'a', 'b', 'c', 'f', 'g']

    f_finite(regexEnumerator, possibilities)


def test_literal_in_character_class():
    regexEnumerator = RegexEnumerator(r'[.]')
    possibilities = ['.']

    f_finite(regexEnumerator, possibilities)


def test_negated_character_class():
    regexEnumerator = RegexEnumerator(r'[^a]')
    possibilities = [chr(i) for i in range(32, 127) if chr(i) != 'a']

    f_finite(regexEnumerator, possibilities)


def test_character_class_with_escaped_special_char_at_start():
    regexEnumerator = RegexEnumerator(r'[\]-a]')
    possibilities = [chr(i) for i in range(93, 98)]

    f_finite(regexEnumerator, possibilities)


def test_character_class_with_escaped_special_char_at_end():
    regexEnumerator = RegexEnumerator(r'[Z-\]]')
    possibilities = [chr(i) for i in range(90, 94)]

    f_finite(regexEnumerator, possibilities)


def test_character_class_with_escape_sequence():
    regexEnumerator = RegexEnumerator(r'[\d]')
    possibilities = [str(i) for i in range(10)]

    f_finite(regexEnumerator, possibilities)


def test_incomplete_range_character_class():
    regexEnumerator = RegexEnumerator(r'[a-]')
    possibilities = ['a', '-']

    f_finite(regexEnumerator, possibilities)


def test_2_ranges():
    regexEnumerator = RegexEnumerator(r'[1a-crf-g3]')
    possibilities = ['1', 'a', 'b', 'c', 'f', 'g', 'r', '3']

    f_finite(regexEnumerator, possibilities)


def test_unicode_character_class():
    regexEnumerator = RegexEnumerator(r'[à-å]')
    possibilities = ['à', 'á', 'â', 'ã', 'ä', 'å']

    f_finite(regexEnumerator, possibilities)


def test_additional_charset():
    regexEnumerator = RegexEnumerator(
        r'[^\w\d\s]', additional_charset=['γ', 'β', 'α'])
    possibilities = ['!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':',
                     ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '`', '{', '|', '}', '~', 'α', 'β', 'γ']

    f_finite(regexEnumerator, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite


def test_digit_escape():
    regexEnumerator = RegexEnumerator(r'\d')
    possibilities = [str(i) for i in range(10)]

    f_finite(regexEnumerator, possibilities)


def test_digit_escape_with_quantifier():
    regexEnumerator = RegexEnumerator(r'\d{1 , 2 }')
    possibilities = [str(i) for i in range(10)] + [str(i) + str(j)
                                                   for i in range(10) for j in range(10)]

    f_finite(regexEnumerator, possibilities)


def test_non_digit_escape():
    regexEnumerator = RegexEnumerator(r'\D')
    possibilities = [chr(i)
                     for i in range(32, 127) if chr(i) not in '0123456789']

    f_finite(regexEnumerator, possibilities)


def test_word_escape():
    regexEnumerator = RegexEnumerator(r'\w')
    possibilities = [chr(i) for i in range(
        32, 127) if chr(i).isalnum() or chr(i) == '_']

    f_finite(regexEnumerator, possibilities)


def test_non_word_escape():
    regexEnumerator = RegexEnumerator(r'\W')
    possibilities = [chr(i) for i in range(
        32, 127) if not (chr(i).isalnum() or chr(i) == '_')]

    f_finite(regexEnumerator, possibilities)


def test_whitespace_escape():
    regexEnumerator = RegexEnumerator(r'\s')
    possibilities = [' ', '\t', '\n', '\r', '\f', '\v']

    f_finite(regexEnumerator, possibilities)


def test_non_whitespace_escape():
    regexEnumerator = RegexEnumerator(r'\S')
    possibilities = [chr(i) for i in range(
        32, 127) if chr(i) not in ' \t\n\r\f\v']

    f_finite(regexEnumerator, possibilities)


def test_tab_escape():
    regexEnumerator = RegexEnumerator(r'\t')
    possibilities = ['\t']

    f_finite(regexEnumerator, possibilities)


def test_carriage_return_escape():
    regexEnumerator = RegexEnumerator(r'\r')
    possibilities = ['\r']

    f_finite(regexEnumerator, possibilities)


def test_newline_escape():
    regexEnumerator = RegexEnumerator(r'\n')
    possibilities = ['\n']

    f_finite(regexEnumerator, possibilities)


def test_vertical_tab_escape():
    regexEnumerator = RegexEnumerator(r'\v')
    possibilities = ['\v']

    f_finite(regexEnumerator, possibilities)


def test_form_feed_escape():
    regexEnumerator = RegexEnumerator(r'\f')
    possibilities = ['\f']

    f_finite(regexEnumerator, possibilities)


def test_hex_escape():
    regexEnumerator = RegexEnumerator(r'\x41')
    possibilities = ['A']

    f_finite(regexEnumerator, possibilities)


def test_escaped_open_square_bracket():
    regexEnumerator = RegexEnumerator(r'\[')
    possibilities = ['[']

    f_finite(regexEnumerator, possibilities)


def test_escaped_open_close_square_brackets():
    regexEnumerator = RegexEnumerator(r'\[\]')
    possibilities = ['[]']

    f_finite(regexEnumerator, possibilities)


def test_escaped_characters_inside_character_class():
    regexEnumerator = RegexEnumerator(r'[\[\]]')
    possibilities = ['[', ']']

    f_finite(regexEnumerator, possibilities)


def test_escaped_char_interrups_range_after_divider():
    regexEnumerator = RegexEnumerator(r'[a-\d]')
    possibilities = ['a', '-', '0', '1', '2',
                     '3', '4', '5', '6', '7', '8', '9']

    f_finite(regexEnumerator, possibilities)


def test_escaped_char_interrups_range_after_1st_char():

    regexEnumerator = RegexEnumerator(r'[\[\d]')
    possibilities = ['[', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

    f_finite(regexEnumerator, possibilities)


def test_escaped_unicode_literal():
    regexEnumerator = RegexEnumerator(r'\u00E0')
    possibilities = ['à']

    f_finite(regexEnumerator, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite


def test_character_class_between_literals():
    regexEnumerator = RegexEnumerator(r'a[0-9]b')
    possibilities = ['a0b', 'a1b', 'a2b', 'a3b',
                     'a4b', 'a5b', 'a6b', 'a7b', 'a8b', 'a9b']

    f_finite(regexEnumerator, possibilities)


def test_single_wildcard():
    regexEnumerator = RegexEnumerator(r'.')
    possibilities = [chr(i) for i in range(32, 127)]

    f_finite(regexEnumerator, possibilities)


def test_wildcard_with_unicode():
    regexEnumerator = RegexEnumerator(r'.', additional_charset='¡¢£')
    possibilities = [chr(i) for i in range(32, 127)] + ['¡', '¢', '£']


def test_done():
    regexEnumerator = RegexEnumerator(r'')
    possibilities = ['', None]

    f_finite(regexEnumerator, possibilities)


def test_weak_password():
    regexEnumerator = RegexEnumerator(
        r'[Ll][Oo0][Vv][Ee3]([Yy][Oo0][Uu])?(2023|2024)[!1.]{1,2}')
    possibilities = []

    you_or_not = []
    for y in 'Yy':
        for o in 'Oo0':
            for u in 'Uu':
                you_or_not.append(y + o + u)
    you_or_not.append('')

    for l_char in 'Ll':
        for o in 'Oo0':
            for v in 'Vv':
                for e in 'Ee3':
                    for y in you_or_not:
                        for year in ['2023', '2024']:
                            for special_1 in ['!', '1', '.']:
                                for special_2 in ['!', '1', '.', '']:
                                    possibilities.append(
                                        l_char + o + v + e + y + year + special_1 + special_2)

    f_finite(regexEnumerator, possibilities)



from .regex_tree import Alternative, BackReference, CharClasses, RegexTree


class RegexError(Exception):
    def __init__(self, regex: str, index: int, message: str):
        self.regex = regex
        self.index = index
        self.message = message

    def __str__(self):
        caret_line = ' ' * self.index + '^'
        return f"\n{self.regex}\n{caret_line}\n{self.message}"


class RegexParser:
    WORDS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
    HEX = '0123456789abcdefABCDEF'
    DIGITS = '0123456789'
    SPACES = ' \t\n\r\f\v'

    def __init__(self, regex: str, charset: str):
        self.regex = regex
        self.charset = charset

    def parse(self) -> RegexTree:
        self.index = 0
        return self._parseRegex(False)

    def _parseRegex(self, to_close: bool) -> RegexTree:
        alternatives: list[Alternative] = []
        elements: list[CharClasses | RegexTree | BackReference] = []
        named_groups: dict[str, RegexTree] = {}
        ordered_groups: list[RegexTree] = []
        min_len_group, max_len_group = 1, 1

        while self.index < len(self.regex):
            char = self.regex[self.index]
            self.index += 1
            match char:
                case'(':
                    if self.index < len(self.regex) and self.regex[self.index] == '?':
                        self.index += 1
                        if self.index >= len(self.regex):
                            self._raise_error("Invalid named group")
                        elif self.regex[self.index] == '<':
                            self.index += 1
                            name = ''
                            while self.index < len(self.regex) and self.regex[self.index] != '>':
                                name += self.regex[self.index]
                                self.index += 1
                            if self.index >= len(self.regex) or self.regex[self.index] != '>' or name == '':
                                self._raise_error("Invalid named group")
                            self.index += 1
                            if name in named_groups:
                                self._raise_error("Duplicate named group")
                            subTree = self._parseRegex(True)
                            named_groups[name] = subTree
                            ordered_groups.append(subTree)
                        elif self.regex[self.index] == ':':
                            self.index += 1
                            subTree = self._parseRegex(True)
                        else:
                            self._raise_error("Invalid group")
                    else:
                        subTree = self._parseRegex(True)
                        ordered_groups.append(subTree)
                    elements.append(subTree)
                case ')':
                    if not to_close:
                        self._raise_error("Unmatched closing parenthesis")
                    min_len_group, max_len_group = self._parseQuantifier()
                    to_close = False
                    break
                case '|':
                    alternatives.append(Alternative(elements))
                    elements = []
                    named_groups = {}
                    ordered_groups = []
                case '[':
                    chars = self._parseCharClass()
                    min_len, max_len = self._parseQuantifier()
                    elements.append(CharClasses(chars, min_len, max_len))
                case '.':
                    chars = list(self.charset)
                    min_len, max_len = self._parseQuantifier()
                    elements.append(CharClasses(chars, min_len, max_len))
                case '\\':
                    reference = self._parseBackReferenceLookahead()
                    if reference is None:
                        chars = self._parseEscapeChar()
                        min_len, max_len = self._parseQuantifier()
                        elements.append(CharClasses(chars, min_len, max_len))
                        continue
                    if isinstance(reference, str):
                        if reference not in named_groups:
                            self._raise_error("Invalid back reference")
                        group = named_groups[reference]
                    else:
                        if reference < 1 or reference > len(ordered_groups):
                            self._raise_error("Invalid back reference")
                        group = ordered_groups[reference - 1]
                    min_len, max_len = self._parseQuantifier()
                    reference = BackReference(group, min_len, max_len)
                    group.add_reference(reference)
                    elements.append(reference)
                case _:
                    min_len, max_len = self._parseQuantifier()
                    elements.append(CharClasses([char], min_len, max_len))

        if to_close:
            self._raise_error("Unmatched opening parenthesis")

        alternatives.append(Alternative(elements))
        return RegexTree(alternatives, min_len_group, max_len_group)

    def _parseBackReferenceLookahead(self) -> str | int | None:
        if len(self.regex) <= self.index:
            self._raise_error("Incomplete escape sequence")

        char = self.regex[self.index]

        match char:
            case 'k':
                self.index += 1
                name = ''
                if len(self.regex) <= self.index or self.regex[self.index] != '<':
                    self._raise_error("Invalid back reference")
                self.index += 1
                while self.index < len(self.regex) and self.regex[self.index] != '>':
                    name += self.regex[self.index]
                    self.index += 1
                if len(self.regex) <= self.index or self.regex[self.index] != '>':
                    self._raise_error("Invalid back reference")
                self.index += 1
                return name
            case char if char.isdigit():
                num = int(char)
                self.index += 1
                while self.index < len(self.regex) and self.regex[self.index].isdigit():
                    num = num * 10 + int(self.regex[self.index])
                    self.index += 1
                return num

    def _parseEscapeChar(self) -> str:

        if len(self.regex) <= self.index:
            self._raise_error("Incomplete escape sequence")

        char = self.regex[self.index]
        self.index += 1

        match char:
            case 'd': return self.DIGITS
            case 'D': return ''.join([c for c in self.charset if not c.isdigit()])
            case 'w': return self.WORDS
            case 'W': return ''.join([c for c in self.charset if c not in self.WORDS])
            case 's': return self.SPACES
            case 'S': return ''.join([c for c in self.charset if c not in self.SPACES])
            case 't': return '\t'
            case 'r': return '\r'
            case 'n': return '\n'
            case 'v': return '\v'
            case 'f': return '\f'
            case 'x':
                if len(self.regex) < self.index + 1 or self.regex[self.index] not in self.HEX:
                    raise ValueError('Invalid escape character')
                if len(self.regex) < self.index + 2 or self.regex[self.index + 1] not in self.HEX:
                    num = int(self.regex[self.index], 16)
                    self.index += 1
                else:
                    num = int(self.regex[self.index: self.index + 2], 16)
                    self.index += 2
                if num < 32 or num > 126:
                    self._raise_error(f"Invalid escape character {num}")
                return chr(num)
            case 'u':
                code = []
                for _ in range(4):
                    if len(self.regex) <= self.index or self.regex[self.index] not in self.HEX:
                        self._raise_error("Invalid escape character")
                    code.append(self.regex[self.index])
                    self.index += 1
                num = int(''.join(code), 16)
                return chr(num)
            case 'p' | 'P':
                self._raise_error("Unicode property not supported")
            case _: return char

    def _parseCharClass(self) -> list[str]:
        chars_list: list[str] = []
        first_char = None
        range_divider = False
        negated = False

        if len(self.regex) <= self.index:
            self._raise_error("Unclosed character class")

        if self.regex[self.index] == '^':
            negated = True
            self.index += 1

        len_regex = len(self.regex)

        while self.index < len_regex and self.regex[self.index] != ']':
            char = self.regex[self.index]
            self.index += 1

            if char == '-' and first_char is not None and not range_divider:
                range_divider = True
                continue
            if char == '\\':
                escape_char = self._parseEscapeChar()
                if len(escape_char) > 1 or escape_char == '-':
                    chars_list.append(escape_char)
                    if range_divider:
                        chars_list.append('-')
                        assert first_char is not None
                        chars_list.append(first_char)
                    elif first_char is not None:
                        chars_list.append(first_char)
                    continue
                char = escape_char

            if first_char is None:
                first_char = char
            elif range_divider:
                chars_list.extend([chr(c) for c in range(
                    ord(first_char), ord(char) + 1)])
                first_char = None
                range_divider = False
            else:
                chars_list.append(first_char)
                first_char = char

        if len(self.regex) <= self.index or self.regex[self.index] != ']':
            self._raise_error("Unclosed character class")

        self.index += 1

        if range_divider:
            chars_list.append('-')
            assert first_char is not None
            chars_list.append(first_char)
        elif first_char is not None:
            chars_list.append(first_char)

        if negated:
            chars_list = [
                c for c in self.charset if c not in ''.join(chars_list)]

        return chars_list

    def _parseQuantifier(self) -> tuple[int, int]:

        if len(self.regex) <= self.index:
            return 1, 1

        char = self.regex[self.index]

        match char:
            case '*':
                self.index += 1
                return 0, None
            case '+':
                self.index += 1
                return 1, None
            case '?':
                self.index += 1
                return 0, 1
            case '{':
                self.index += 1
                return self._parseMinMax()
            case _: return 1, 1

    def _parseMinMax(self) -> tuple[int, int | None]:
        self._skipSpaces()

        min_len = 0
        if self.index >= len(self.regex) or not self.regex[self.index].isdigit():
            self._raise_error("Invalid quantifier")
        while self.index < len(self.regex) and self.regex[self.index].isdigit():
            min_len = min_len * 10 + int(self.regex[self.index])
            self.index += 1

        self._skipSpaces()

        if self.index >= len(self.regex):
            self._raise_error("Invalid quantifier")

        if self.regex[self.index] == '}':
            self.index += 1
            return min_len, min_len
        if self.regex[self.index] != ',':
            self._raise_error("Invalid quantifier")

        self.index += 1
        self._skipSpaces()

        if self.index >= len(self.regex) or self.regex[self.index] not in '0123456789}':
            self._raise_error("Invalid quantifier")

        if self.regex[self.index] == '}':
            self.index += 1
            return min_len, None

        max_len = 0
        while self.index < len(self.regex) and self.regex[self.index].isdigit():
            max_len = max_len * 10 + int(self.regex[self.index])
            self.index += 1

        if max_len < min_len:
            self._raise_error(
                "Max length cannot be less than min length in quantifier")

        self._skipSpaces()

        if self.index >= len(self.regex) or self.regex[self.index] != '}':
            self._raise_error("Invalid quantifier")
        self.index += 1

        return min_len, max_len

    def _skipSpaces(self):
        while self.index < len(self.regex) and self.regex[self.index] == ' ':
            self.index += 1

    def _raise_error(self, message: str):
        raise RegexError(self.regex, self.index, message)



class RegexTree:
    pass


class CharClasses:
    def __init__(self, chars_list: list[str], min_len: int, max_len: int):
        self._index = 0
        self._chars: str = ''.join(sorted(set(''.join(chars_list))))
        self._min_len = min_len
        self._max_len = max_len
        self._base = len(self._chars)
        self.done = self._base == 0 or self._max_len == 0
        self._max_index = self._calculate_max_index()
        self.current: set[str] = {self._calculate()} if not self.done else {''}

    def _calculate_max_index(self) -> int | None:
        if self._max_len is None or self.done:
            return None
        if self._base == 1:
            return self._max_len - self._min_len
        return ((self._base ** self._min_len - self._base ** (self._max_len + 1)) // (1 - self._base)) - 1

    def _calculate(self) -> str:
        if self._max_len is not None and self._index >= self._max_index:
            self.done = True

        if self._base == 1:
            return self._chars[0] * (self._min_len + self._index)

        res = []
        num = self._base ** self._min_len + self._index
        while num > 1:
            res.append(self._chars[num % self._base])
            num //= self._base

        return ''.join(reversed(res))

    def next(self) -> set[str]:
        assert not self.done

        self._index += 1
        res = self._calculate()
        self.current.add(res)
        return {res}


class BackReference:
    def __init__(self, reference: RegexTree, min_len: int, max_len: int | None):
        self._min_len = min_len
        self._max_len = max_len
        self._index = 0
        self.reference: RegexTree = reference
        self.done = max_len == 0 or (
            reference.done and len(reference.current) == 0)
        self.current: dict[str, list[str]
                           ] = self._calculate() if not self.done else {}

    def update(self):
        if self._max_len is not None and self._min_len + self._index >= self._max_len and self.reference.done:
            self.done = True

        for string in self.reference.current:
            if string in self.current:
                self.current[string].append(
                    string * (self._min_len + self._index))
            else:
                self.current[string] = []
                for i in range(self._min_len, self._min_len + self._index + 1):
                    self.current[string].append(string * i)

    def _calculate(self) -> dict[str, set[str]]:
        current_ref = self.reference.current
        if self._max_len is not None and self._min_len + self._index >= self._max_len:
            self.done = True

        result: dict[str, list[str]] = {}

        for string in current_ref:
            result[string] = [string * (self._min_len + self._index)]

        return result

    def next(self) -> dict[str, set[str]]:
        assert not self.done
        self._index += 1
        if self._max_len is not None and self._min_len + self._index >= self._max_len:
            self.done = True

        for string in self.current.keys():
            self.current[string].append(string * (self._min_len + self._index))

        return self.current


class Alternative:
    def __init__(self, elements: list[CharClasses | RegexTree | BackReference]):
        self._index = 0
        self._elements: list[CharClasses | RegexTree | BackReference] = [
            element for element in elements if not element.done or len(element.current) > 0]
        self._base = len(self._elements)
        self.done = self._base == 0
        self.current: set[str] = self._calculate() if not self.done else {''}

    def next(self) -> set[str]:
        assert not self.done
        assert not isinstance(self._elements[0], BackReference)

        self._index += 1
        for i in range(self._index, self._index + self._base):
            if self._elements[i % self._base].done:
                self._index += 1
            else:
                break

        result: list[tuple[str, dict[RegexTree, str]]] = []

        if isinstance(self._elements[0], CharClasses):
            for string in self._elements[0].next() if self._index % self._base == 0 else self._elements[0].current:
                result.append((string, {}))
        else:
            for string in self._elements[0].next() if self._index % self._base == 0 else self._elements[0].current:
                result.append((string, {self._elements[0]: string}))

        done = self._elements[0].done

        for i, element in enumerate(self._elements[1:], start=1):
            temp = []
            if isinstance(element, CharClasses):
                for sfx in element.next() if i == self._index % self._base else element.current:
                    for pfx in result:
                        temp.append((pfx[0] + sfx, pfx[1]))
            elif isinstance(element, RegexTree):
                for sfx in element.next() if i == self._index % self._base else element.current:
                    for pfx in result:
                        temp.append((pfx[0] + sfx, {**pfx[1], element: sfx}))
            else:
                if i == self._index % self._base:
                    element.next()
                for pfx in result:
                    reference = pfx[1][element.reference]
                    assert reference is not None
                    for sfx in element.current[reference]:
                        temp.append(
                            (pfx[0] + sfx, pfx[1]))
            result = temp
            done = done and element.done

        self.done = done
        new_res = {struct[0] for struct in result} - self.current
        self.current.update(new_res)
        return new_res

    def _calculate(self) -> set[str]:
        assert not isinstance(self._elements[0], BackReference)

        result: list[tuple[str, dict[RegexTree, str]]] = []

        if isinstance(self._elements[0], CharClasses):
            for char in self._elements[0].current:
                result.append((char, {}))
        else:
            for char in self._elements[0].current:
                result.append((char, {self._elements[0]: char}))

        done = self._elements[0].done

        for element in self._elements[1:]:
            temp: list[tuple[str, dict[RegexTree, str]]] = []
            done = done and element.done
            if isinstance(element, CharClasses):
                for pfx in result:
                    for sfx in element.current:
                        temp.append((pfx[0] + sfx, pfx[1]))
            elif isinstance(element, RegexTree):
                for pfx in result:
                    for sfx in element.current:
                        temp.append((pfx[0] + sfx, {**pfx[1], element: sfx}))
            else:
                for pfx in result:
                    reference = pfx[1][element.reference]
                    assert reference is not None
                    for sfx in element.current[reference]:
                        temp.append(
                            (pfx[0] + sfx, pfx[1]))
            result = temp

        self.done = done
        return {struct[0] for struct in result}


class RegexTree:
    def __init__(self, alternatives: list[Alternative], min_len: int, max_len: int | None):
        self.references: list[BackReference] = []
        self._alternatives: list[Alternative] = [
            alternative for alternative in alternatives if not alternative.done or len(alternative.current) > 0]
        self._min_len = min_len
        self._max_len = max_len
        self._base = len(self._alternatives)
        self.done = self._base == 0 or self._max_len == 0
        self._gen_charset = False
        self._index_charset = 0
        self._index_repetition = 0
        self._done_repetition = False
        self._current_chars: set[str] = self._calculate_chars()
        self.current: set[str] = self._calculate() if not self.done else set()

    def add_reference(self, reference: BackReference):
        if reference.done and len(reference.current) == 0:
            return
        self.references.append(reference)

    def _calculate_using_new_charset(self) -> set[str]:
        assert not self.done
        if self._done_repetition and self._done_charset:
            self.done = True

        if self._index_repetition + self._min_len == 0:
            return {''}

        result = set(self._current_chars)
        for _ in range(1, self._min_len + self._index_repetition):
            result.update({pfx + sfx for pfx in result for sfx in self._current_chars})
                
        return result

    def next(self) -> set[str]:
        assert not self.done

        if self._done_charset:
            self._gen_charset = False
        elif self._done_repetition:
            self._gen_charset = True

        if self._gen_charset:
            _: set[str] = self._next_charset()
            # Optimization: use the new charset to calculate the next set of strings
            res: set[str] = self._calculate_using_new_charset()
        else:
            if not self._done_repetition:
                self._index_repetition += 1
            res: set[str] = self._calculate()

        self._gen_charset = not self._gen_charset
        new_res = res - self.current
        if len(new_res) == 0:
            return new_res

        self.current.update(new_res)
        if len(self.references) == 0:
            return new_res

        for reference in self.references:
            reference.update()
        return self.current

    def _calculate(self) -> set[str]:
        if self._max_len is not None and self._index_repetition + self._min_len >= self._max_len:
            self._done_repetition = True
            if self._done_charset:
                self.done = True

        if self._index_repetition + self._min_len == 0:
            return {''}

        result = set(self._current_chars)
        for _ in range(1, self._min_len + self._index_repetition):
            result = {pfx + sfx for pfx in result for sfx in self._current_chars}

        return result

    def _next_charset(self) -> set[str]:
        assert not self._done_charset

        self._index_charset += 1

        for i in range(self._index_charset, self._index_charset + self._base):
            if self._alternatives[i % self._base].done:
                self._index_charset += 1
            else:
                break

        res = self._alternatives[self._index_charset % self._base].next()
        done_charset = True

        for alternative in self._alternatives:
            if not alternative.done:
                done_charset = False
                break

        new_res = res - self._current_chars
        self._current_chars.update(new_res)
        self._done_charset = done_charset
        return new_res

    def _calculate_chars(self) -> set[str]:
        res = set()
        done_charset = True

        for alternative in self._alternatives:
            res.update(alternative.current)
            done_charset = done_charset and alternative.done

        self._done_charset = done_charset
        return res



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_single_capturing_group_with_literal():
    regexEnumerator = RegexEnumerator(r'(a)')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)


def test_single_capturing_group_with_class_single_char():
    regexEnumerator = RegexEnumerator(r'([a])')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)


def test_single_capturing_group_with_class_multi_char():
    regexEnumerator = RegexEnumerator(r'([a-c])')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)


def test_capturing_group_with_star_quantifier():
    regexEnumerator = RegexEnumerator(r'(a)*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)


def test_named_capturing_group_with_optional_subgroup():
    regexEnumerator = RegexEnumerator(r'(?<name>a[bcd](e)?)')
    possibilities = ['ab', 'abe', 'ac', 'ace', 'ad', 'ade']

    f_finite(regexEnumerator, possibilities)


def test_literal_followed_by_group_with_star_quantifier():
    regexEnumerator = RegexEnumerator(r'a(b)*')
    possibilities = ['a' + 'b' * i for i in range(6)]

    f_infinite(regexEnumerator, possibilities)


def test_two_capturing_groups_with_star_quantifiers():
    regexEnumerator = RegexEnumerator(r'(a)*(b)*')
    possibilities = ['a' * i + 'b' * j for i in range(6) for j in range(6)]

    f_infinite(regexEnumerator, possibilities)


def test_nested_capturing_groups():
    regexEnumerator = RegexEnumerator(r'(a(b(c)))')
    possibilities = ['abc']

    f_finite(regexEnumerator, possibilities)


def test_capturing_groups_in_sequence():
    regexEnumerator = RegexEnumerator(r'((a)(b))')
    possibilities = ['ab']

    f_finite(regexEnumerator, possibilities)


def test_non_capturing_group():
    regexEnumerator = RegexEnumerator(r'(?:a|b)*')
    possibilities = ['', 'a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_infinite(regexEnumerator, possibilities)


def test_non_capturing_group_with_quantifier():
    regexEnumerator = RegexEnumerator(r'(?:ab)+')
    possibilities = ['ab', 'abab', 'ababab']

    f_infinite(regexEnumerator, possibilities)


def test_named_capturing_group_with_quantifier():
    regexEnumerator = RegexEnumerator(r'(?<chars>[ab]{1,2})')
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_finite(regexEnumerator, possibilities)


def test_nested_non_capturing_groups():
    regexEnumerator = RegexEnumerator(r'(?:a(?:b(?:c)))?')
    possibilities = ['', 'abc']

    f_finite(regexEnumerator, possibilities)


def test_group_for_quantifier_scope():
    regexEnumerator = RegexEnumerator(r'(ab)+')
    possibilities = ['ab', 'abab', 'ababab']

    f_infinite(regexEnumerator, possibilities)

def test_group_with_char_class_infinite_repetition():
    regexEnumerator = RegexEnumerator(r'([ab])+')
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_infinite(regexEnumerator, possibilities)


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='regex_enumerator',
    version='0.8.2',
    packages=find_packages(include=['regex_enumerator', 'regex_enumerator.*']),
    description='Enumerate all strings that match a given regex',
    author='Vincenzo Greco',
    author_email='grecovincenzo98@gmail.com',
    extras_require={
        'dev': ['pytest', 'pytest-cov'],
    },
    url='https://github.com/Buba98/regex_enumerator',
    keywords=['regex', 'regex enumerator', 'regular-expression', 'enumerator', 'string-generation',
              'exhaustive-matching', 'regex-testing', 'regex-tools', 'string-enumeration', 'data-generation'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.10',
)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_backreference():
    regexEnumerator = RegexEnumerator(r'(a)\1')
    possibilities = ['aa']

    f_finite(regexEnumerator, possibilities)


def test_backreference_with_group_quantifier():
    regexEnumerator = RegexEnumerator(r'(a)+\1')
    possibilities = ['aa' * i for i in range(1, 6)]

    f_infinite(regexEnumerator, possibilities)


def test_backreference_with_quantifier():
    regexEnumerator = RegexEnumerator(r'(a)\1+')
    possibilities = ['a' * i + 'a' for i in range(1, 6)]

    f_infinite(regexEnumerator, possibilities)


def test_backreference_with_named_group():
    regexEnumerator = RegexEnumerator(r'(?<name>[a-b])\k<name>')
    possibilities = ['aa', 'bb']

    f_finite(regexEnumerator, possibilities)


def test_backreference_with_named_group_and_quantifier():
    regexEnumerator = RegexEnumerator(r'(?<name>[a-b])\k<name>{1, 2}')
    possibilities = ['aa', 'bb', 'aaa', 'bbb']

    f_finite(regexEnumerator, possibilities)


def test_zero_width_backreference():
    regexEnumerator = RegexEnumerator(r'(a)?\1{0}')
    possibilities = ['a', '']

    f_finite(regexEnumerator, possibilities)


def test_10_backreference():
    regexEnumerator = RegexEnumerator(r'(a)(b)(c)(d)(e)(f)(g)(h)(i)(j)\10')
    possibilities = ['abcdefghijj']

    f_finite(regexEnumerator, possibilities)


def test_multiple_backreferences():
    regexEnumerator = RegexEnumerator(r'(a)(b)\2\1')
    possibilities = ['abba']

    f_finite(regexEnumerator, possibilities)


def test_backreference_with_mismatch():
    regexEnumerator = RegexEnumerator(r'(a)(b)\1')
    possibilities = ['aba']

    f_finite(regexEnumerator, possibilities)


def test_named_group_with_backreference():
    regexEnumerator = RegexEnumerator(r'(?<letter>[ab])\k<letter>')
    possibilities = [
        'aa', 'bb'
    ]

    f_finite(regexEnumerator, possibilities)


def test_named_group_infinite_repetition_with_backreference():
    regexEnumerator = RegexEnumerator(r'(?<letter>[ab])+\k<letter>')
    possibilities = [
        'aa', 'bb', 'abab', 'baba', 'aaaa', 'bbbb'
    ]

    f_infinite(regexEnumerator, possibilities)